=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-10-02 21:51:29 +0000
+++ bzrlib/errors.py	2005-10-05 10:02:53 +0000
@@ -27,7 +27,10 @@
             return self.args[0]
         elif len(self.args) == 2:
             # further explanation or suggestions
-            return '\n  '.join([self.args[0]] + self.args[1])
+            try:
+                return '\n  '.join([self.args[0]] + self.args[1])
+            except TypeError:
+                return "%r" % self
         else:
             return `self.args`
 
@@ -203,6 +206,10 @@
 
 class NoSuchFile(TransportError, IOError):
     """A get() was issued for a file that doesn't exist."""
+
+    def __str__(self):
+        return self.msg
+
     def __init__(self, msg=None, orig_error=None):
         import errno
         TransportError.__init__(self, msg=msg, orig_error=orig_error)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-10-05 06:44:02 +0000
+++ bzrlib/fetch.py	2005-10-05 10:02:53 +0000
@@ -82,7 +82,7 @@
 
     count_copied -- number of revisions copied
 
-    count_texts -- number of file texts copied
+    count_weaves -- number of file weaves copied
     """
     def __init__(self, to_branch, from_branch, last_revision=None, pb=None):
         if to_branch == from_branch:
@@ -96,7 +96,7 @@
         self.failed_revisions = []
         self.count_copied = 0
         self.count_total = 0
-        self.count_texts = 0
+        self.count_weaves = 0
         self.copied_file_ids = set()
         if pb is None:
             self.pb = bzrlib.ui.ui_factory.progress_bar()
@@ -212,17 +212,15 @@
         # in memory until everything's done?  But this way is nicer
         # if it's interrupted.
         for path, ie in inv.iter_entries():
-            if not ie.has_text():
-                continue
             if ie.revision != rev_id:
                 continue
             mutter('%s {%s} is changed in this revision',
                    path, ie.file_id)
-            self._copy_one_text(rev_id, ie.file_id)
-
-
-    def _copy_one_text(self, rev_id, file_id):
-        """Copy one file text."""
+            self._copy_one_weave(rev_id, ie.file_id)
+
+
+    def _copy_one_weave(self, rev_id, file_id):
+        """Copy one file weave."""
         mutter('copy file {%s} modified in {%s}', file_id, rev_id)
         if file_id in self.copied_file_ids:
             mutter('file {%s} already copied', file_id)
@@ -231,7 +229,7 @@
         to_weave = self.to_weaves.get_weave_or_empty(file_id)
         to_weave.join(from_weave)
         self.to_weaves.put_weave(file_id, to_weave)
-        self.count_texts += 1
+        self.count_weaves += 1
         self.copied_file_ids.add(file_id)
         mutter('copied file {%s}', file_id)
 

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2005-10-04 06:39:39 +0000
+++ bzrlib/inventory.py	2005-10-05 10:02:53 +0000
@@ -159,6 +159,10 @@
         """Return true if the object this entry represents has textual data.
 
         Note that textual data includes binary content.
+
+        Also note that all entries get weave files created for them.
+        This attribute is primarily used when upgrading from old trees that
+        did not have the weave index for all inventory entries.
         """
         return False
 
@@ -276,7 +280,7 @@
         if len(previous_entries) == 1:
             # cannot be unchanged unless there is only one parent file rev.
             parent_ie = previous_entries.values()[0]
-            if self._unchanged(path, parent_ie, work_tree):
+            if self._unchanged(parent_ie):
                 mutter("found unchanged entry")
                 self.revision = parent_ie.revision
                 return "unchanged"
@@ -289,25 +293,17 @@
         mutter('new revision for {%s}', self.file_id)
         self.revision = revision
         change = self._get_snapshot_change(previous_entries)
+        self._snapshot_text(previous_entries, work_tree, weave_store)
         return change
 
     def _snapshot_text(self, file_parents, work_tree, weave_store): 
+        """Record the 'text' of this entry, whatever form that takes.
+        
+        This default implementation simply adds an empty text.
+        """
         mutter('storing file {%s} in revision {%s}',
                self.file_id, self.revision)
-        # special case to avoid diffing on renames or 
-        # reparenting
-        if (len(file_parents) == 1
-            and self.text_sha1 == file_parents.values()[0].text_sha1
-            and self.text_size == file_parents.values()[0].text_size):
-            previous_ie = file_parents.values()[0]
-            weave_store.add_identical_text(
-                self.file_id, previous_ie.revision, 
-                self.revision, file_parents)
-        else:
-            new_lines = work_tree.get_file(self.file_id).readlines()
-            self._add_text_to_weave(new_lines, file_parents, weave_store)
-            self.text_sha1 = sha_strings(new_lines)
-            self.text_size = sum(map(len, new_lines))
+        self._add_text_to_weave([], file_parents, weave_store)
 
     def __eq__(self, other):
         if not isinstance(other, InventoryEntry):
@@ -331,7 +327,7 @@
     def __hash__(self):
         raise ValueError('not hashable')
 
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _unchanged(self, previous_ie):
         """Has this entry changed relative to previous_ie.
 
         This method should be overriden in child classes.
@@ -502,18 +498,29 @@
         self.text_sha1 = work_tree.get_file_sha1(self.file_id)
         self.executable = work_tree.is_executable(self.file_id)
 
-    def snapshot_revision(self, revision, previous_entries, work_tree,
-                          weave_store):
-        """See InventoryEntry.snapshot_revision."""
-        change = super(InventoryFile, self).snapshot_revision(revision, 
-            previous_entries, work_tree, weave_store)
-        self._snapshot_text(previous_entries, work_tree, weave_store)
-        return change
-
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _snapshot_text(self, file_parents, work_tree, weave_store): 
+        """See InventoryEntry._snapshot_text."""
+        mutter('storing file {%s} in revision {%s}',
+               self.file_id, self.revision)
+        # special case to avoid diffing on renames or 
+        # reparenting
+        if (len(file_parents) == 1
+            and self.text_sha1 == file_parents.values()[0].text_sha1
+            and self.text_size == file_parents.values()[0].text_size):
+            previous_ie = file_parents.values()[0]
+            weave_store.add_identical_text(
+                self.file_id, previous_ie.revision, 
+                self.revision, file_parents)
+        else:
+            new_lines = work_tree.get_file(self.file_id).readlines()
+            self._add_text_to_weave(new_lines, file_parents, weave_store)
+            self.text_sha1 = sha_strings(new_lines)
+            self.text_size = sum(map(len, new_lines))
+
+
+    def _unchanged(self, previous_ie):
         """See InventoryEntry._unchanged."""
-        compatible = super(InventoryFile, self)._unchanged(path, previous_ie, 
-            work_tree)
+        compatible = super(InventoryFile, self)._unchanged(previous_ie)
         if self.text_sha1 != previous_ie.text_sha1:
             compatible = False
         else:
@@ -597,10 +604,9 @@
         """See InventoryEntry._read_tree_state."""
         self.symlink_target = work_tree.get_symlink_target(self.file_id)
 
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _unchanged(self, previous_ie):
         """See InventoryEntry._unchanged."""
-        compatible = super(InventoryLink, self)._unchanged(path, previous_ie, 
-            work_tree)
+        compatible = super(InventoryLink, self)._unchanged(previous_ie)
         if self.symlink_target != previous_ie.symlink_target:
             compatible = False
         return compatible

=== modified file 'bzrlib/selftest/__init__.py'
--- bzrlib/selftest/__init__.py	2005-10-05 02:47:53 +0000
+++ bzrlib/selftest/__init__.py	2005-10-05 10:02:53 +0000
@@ -380,6 +380,10 @@
                 print >>f, "contents of", name
                 f.close()
 
+    def failUnlessExists(self, path):
+        """Fail unless path, which may be abs or relative, exists."""
+        self.failUnless(os.path.exists(path))
+        
 
 class MetaTestLog(TestCase):
     def test_logging(self):

=== modified file 'bzrlib/selftest/test_upgrade.py'
--- bzrlib/selftest/test_upgrade.py	2005-10-04 11:51:06 +0000
+++ bzrlib/selftest/test_upgrade.py	2005-10-05 10:02:53 +0000
@@ -29,12 +29,14 @@
 from bzrlib.upgrade import upgrade
 from bzrlib.selftest.treeshape import build_tree_contents
 
+
 class TestUpgrade(TestCaseInTempDir):
+    
     def test_build_tree(self):
         """Test tree-building test helper"""
         build_tree_contents(_upgrade1_template)
-        self.assertTrue(os.path.exists('foo'))
-        self.assertTrue(os.path.exists('.bzr/README'))
+        self.failUnlessExists('foo')
+        self.failUnlessExists('.bzr/README')
 
     def test_upgrade_simple(self):
         """Upgrade simple v0.0.4 format to v5"""
@@ -72,6 +74,11 @@
         eq(len(rev.parent_ids), 2)
         eq(rev.parent_ids[1], 'wibble@wobble-2')
 
+    def test_upgrade_makes_dir_weaves(self):
+        build_tree_contents(_upgrade_dir_template)
+        upgrade('.')
+        self.failUnlessExists(
+            '.bzr/weaves/dir-20051005095101-da1441ea3fa6917a.weave')
 
 
 _upgrade1_template = \
@@ -180,3 +187,54 @@
         '\xd0\x9a`\x0c*W\x1aR\xc1\x94du\x08(\t\xb0\x91\xdeZ\xa3\x9cU\x9cm\x7f\x8dr\x1d\x10Ot\xb8\xc6\xcf\xa7\x907|\xfb-\xb1\xbd\xd3\xfb\xd5\x07\xeeD\xee\x08*\x02\x00\x00'
     ),
 ]
+
+_upgrade_dir_template = [
+    ( './.bzr/', ),
+    ( './.bzr/README',
+        'This is a Bazaar-NG control directory.\n'
+        'Do not change any files in this directory.\n'
+    ),
+    ( './.bzr/branch-format',
+        'Bazaar-NG branch, format 0.0.4\n'
+    ),
+    ( './.bzr/branch-lock',
+        ''
+    ),
+    ( './.bzr/branch-name',
+        ''
+    ),
+    ( './.bzr/inventory',
+        '<inventory>\n'
+        '<entry file_id="dir-20051005095101-da1441ea3fa6917a" kind="directory" name="dir" />\n'
+        '</inventory>\n'
+    ),
+    ( './.bzr/merged-patches',
+        ''
+    ),
+    ( './.bzr/pending-merged-patches',
+        ''
+    ),
+    ( './.bzr/pending-merges',
+        ''
+    ),
+    ( './.bzr/revision-history',
+        'robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e\n'
+    ),
+    ( './.bzr/stat-cache',
+        '### bzr hashcache v5\n'
+    ),
+    ( './.bzr/text-store/', ),
+    ( './.bzr/inventory-store/', ),
+    ( './.bzr/inventory-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
+        '\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xb3\xc9\xcc+K\xcd+\xc9/\xaa\xb4\xe3\xb2\x012\x8a*\x15\xd22sR\xe33Sl\x95R2\x8bt\x8d\x0c\x0cL\r'
+        "\x81\xd8\xc0\x12H\x19\xea\xa6$\x1a\x9a\x98\x18\xa6&\x1a\xa7%\x9aY\x1a\x9a'*)dg\xe6A\x94\xa6&\x83LQR\xc8K\xccM\x05\x0b()\xe8\x03\xcd\xd4G\xb2\x00\x00\xc2<\x94\xb1m\x00\x00\x00"
+    ),
+    ( './.bzr/revision-store/', ),
+    ( './.bzr/revision-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
+        '\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xa5OKj\xc30\x14\xdc\xfb\x14B\x8b\xec\x92<I\xd6\xc7\xc42\x85\xde\xa0\x17(\xb6\xf4\x9c\n'
+        'l\xa9H"\x90\x9c\xbe\xa6\xa9\xa1\x9b\xae\xbax\x0c\xcc\xe71\xd3g\xbc\x85\x12R$.\xadk\xa8\x15\xb3\xa5oi\xc2\\\xc9kZ\x96\x10\x0b9,\xf5\x92\xbf)\xf7\xf2\x83O\xe5\x14\xb1\x1e\xae\xf5BI\x887\x8c5\xe5\xfb{\xf0\x96\xfei>r\x00\xc9\xb6\x83n\x03sT\xa0\xe4<y\x83\xda\x1b\xc54\xfe~T>Ff\xe9\xcc:\xdd\x8e\xa6E\xc7@\xa2\x82I\xaaNL\xbas\\313)\x00\xb9\xe6\xe0(\xd9\x87\xfc\xb7A\r'
+        "+\x96:\xae\x9f\x962\xc6\x8d\x04i\x949\x01\x97R\xb7\x1d\x17O\xc3#E\xb4T(\x00\xa0C\xd3o\x892^q\x18\xbd'>\xe4\xfe\xbc\x13M\x7f\xde{\r"
+        '\xcd\x17\x85\xea\xba\x03l\x01\x00\x00'
+    ),
+    ( './dir/', ),
+]

=== modified file 'bzrlib/upgrade.py'
--- bzrlib/upgrade.py	2005-10-04 11:51:06 +0000
+++ bzrlib/upgrade.py	2005-10-05 10:02:53 +0000
@@ -323,76 +323,64 @@
         parent_invs = map(self._load_updated_inventory, present_parents)
         for file_id in inv:
             ie = inv[file_id]
-            self._set_revision(rev, ie, parent_invs)
-            if not ie.has_text():
-                continue
             self._convert_file_version(rev, ie, parent_invs)
 
-
-    def _set_revision(self, rev, ie, parent_invs):
-        """Set name version for a file.
-
-        Done in a slightly lazy way: if the file is renamed or in a merge revision
-        it gets a new version, otherwise the same as before.
-        """
-        file_id = ie.file_id
-        if ie.kind == 'root_directory':
-            return
-        if len(parent_invs) != 1:
-            ie.revision = rev.revision_id
-        else:
-            old_inv = parent_invs[0]
-            if not old_inv.has_id(file_id):
-                ie.revision = rev.revision_id
-            else:
-                old_ie = old_inv[file_id]
-                if (old_ie.parent_id != ie.parent_id
-                    or old_ie.name != ie.name):
-                    ie.revision = rev.revision_id
-                else:
-                    ie.revision = old_ie.revision
-
-
-
     def _convert_file_version(self, rev, ie, parent_invs):
         """Convert one version of one file.
 
         The file needs to be added into the weave if it is a merge
         of >=2 parents or if it's changed from its parent.
         """
+        if ie.kind == 'root_directory':
+            return
         file_id = ie.file_id
         rev_id = rev.revision_id
         w = self.text_weaves.get(file_id)
         if w is None:
             w = Weave(file_id)
             self.text_weaves[file_id] = w
-        file_parents = []
         text_changed = False
+        previous_revisions = {}
         for parent_inv in parent_invs:
             if parent_inv.has_id(file_id):
-                parent_ie = parent_inv[file_id]
-                old_revision = parent_ie.revision
+                previous_ie = parent_inv[file_id]
+                if previous_ie.revision in previous_revisions:
+                    assert previous_revisions[previous_ie.revision] == previous_ie
+                else:
+                    previous_revisions[previous_ie.revision] = previous_ie
+                old_revision = previous_ie.revision
+        for old_revision in previous_revisions:
                 # if this fails, its a ghost ?
                 assert old_revision in self.converted_revs 
-                if old_revision not in file_parents:
-                    file_parents.append(old_revision)
-                if parent_ie.text_sha1 != ie.text_sha1:
-                    text_changed = True
-        if len(file_parents) != 1 or text_changed:
+        self.snapshot_ie(previous_revisions, ie, w, rev_id)
+        del ie.text_id
+        assert getattr(ie, 'revision', None) is not None
+
+    def snapshot_ie(self, previous_revisions, ie, w, rev_id):
+        # TODO: convert this logic, which is ~= snapshot to
+        # a call to:. This needs the path figured out. rather than a work_tree
+        # a v4 revision_tree can be given, or something that can give the
+        # text for the ie if it needs it.
+        # and we need something that looks like a weave store for snapshot to 
+        # save against.
+        #ie.snapshot(rev, PATH, previous_revisions, REVISION_TREE, InMemoryWeaveStore(self.text_weaves))
+        if len(previous_revisions) == 1:
+            previous_ie = previous_revisions.values()[0]
+            if ie._unchanged(previous_ie):
+                ie.revision = previous_ie.revision
+                return
+        parent_indexes = map(w.lookup, previous_revisions)
+        if ie.has_text():
             file_lines = self.branch.text_store[ie.text_id].readlines()
             assert sha_strings(file_lines) == ie.text_sha1
             assert sum(map(len, file_lines)) == ie.text_size
-            w.add(rev_id, file_parents, file_lines, ie.text_sha1)
-            ie.revision = rev_id
+            w.add(rev_id, parent_indexes, file_lines, ie.text_sha1)
             self.text_count += 1
-            ##mutter('import text {%s} of {%s}',
-            ##       ie.text_id, file_id)
         else:
-            ##mutter('text of {%s} unchanged from parent', file_id)
-            ie.revision = file_parents[0]
-        del ie.text_id
-
-
+            w.add(rev_id, parent_indexes, [], None)
+        ie.revision = rev_id
+        ##mutter('import text {%s} of {%s}',
+        ##       ie.text_id, file_id)
 
     def _make_order(self):
         """Return a suitable order for importing revisions.

